/**********************************************************************************************************
*   Suivre les variations du secteur sur un bargraph en mode 3-6-12-24 heures grace à un ATMEGA328P sur un*
* afficheur 4 lignes -20 caractères, en drive parallèle avec extinction du rétro-éclairage. Interruption  *
* active sur bouton poussoir pour effacer les données ou changer la résolution de la base de temps . La   *
* plage de tension est modifiable dans les déclaration et la base de temps également . J'ai repris une    *
* fabrication de caractères spéciaux existante pour les cellules d'affichage du bargraph. En lire plus:   *
*   https://www.amazing-vacuum-tubes.com/pages/pour-y-arriver/voltgraph-maton-pour-le-secteur-.html       *
**********************************************************************************************************/

#include <LiquidCrystal.h>
#include "GyverButton.h"
               
const int LedPin = 19;      // Pin de cde rétro-eclairage
#define BTN_PIN 2
#define BASE_PERIOD 675000  //Période d'acquisition du tableau principal 675 000 ms = 11,25 min (*16 barres = 3 heures) résolution graphique maximale
#define MIN_VAL 210         //La valeur minimale affichée sur le graphique
#define MAX_VAL 240         //La valeur maximale affichée sur le graphique
LiquidCrystal lcd(6, 5, 4, 3, 1, 0);  //Pin LCD: RS,E, D4-D7,RW=0V,
GButton butt1(BTN_PIN);

uint32_t tmr1, tmr2; //variables de Tempo
uint32_t  set_period = BASE_PERIOD; //Période d'affichage, change en multiples de 3 heures (3, 6, 12, 24)
int16_t plot_array[20]; // Tableau de données pour dessiner un graphique
uint16_t base_array [128];  //Le tableau de base stock toutes les mesures des dernières 24 heures (128 cellules * 11,25 minutes = 1440 minutes = 24 heures)
int16_t value, delta; //Lectures actuelles (prises), différence entre les lectures pour une période de temps sélectionnée
byte interval = 1; //Intervalle affiché à l'écran (différence de temps entre les barres de graphique adjacentes)
                   // 11,25 min * 16 = 3 heures, 22,5 min - 6 heures, 45 min - 12 heures, 90 min - 24 heures
void setup() {
  lcd.begin(20,4); //Afficheur 4 lignes 20 caractères
  read_all ();
  attachInterrupt(2, isr, CHANGE);  // Declenche l'interruption sur un changement d'etat
  butt1.setDebounce(80); // paramètre anti-rebond (par défaut 80 ms)
  butt1.setTimeout(300); // paramètre de délai d'attente de maintien enfoncé (par défaut 500 ms)
  
  digitalWrite(LedPin, HIGH); // Allume le rétroéclage de l'écran
    lcd.clear();
    lcd.setCursor(3, 1);
    lcd.print(F("amazing-vacuum-"));  //"C'est Moi"
    lcd.setCursor(5, 2);
    lcd.print(F(" tubes.com"));
    delay (3000);


  if (!digitalRead(BTN_PIN)) { //Réinitialiser les paramètres lors de la mise sous tension avec le bouton maintenu enfoncé                         
    for (byte i = 0; i < 128; i++) base_array [i] = 0;  //Efface les données
    update_all (); //Ecrit des 0 en EEprom
    lcd.setCursor(1, 1);  //Et signaler
    lcd.print(F("Effacement des datas"));
    lcd.setCursor(9, 2);
    lcd.print(F("OK"));
  }
  while (!digitalRead(BTN_PIN));
  lcd.clear();
    
  initPlot(); //Initialisation des symboles pour le rendu visuel                                          
  
  value = round (analogRead (A0))*5/10.23; //Résolution sur 5 volts / 1023 pas * tension lue * 100 = valeur
  base_array[0] = value;
  get_data ();
}

void isr() {  //Interrogation du bouton dans l'interruption pour attraper le clic de toute façon
   butt1.tick();
}

void loop() {
  butt1.tick();  //Test du bouton                                       
  if (butt1.isClick()) {  //quand ile est appuyé:
    digitalWrite(LedPin, HIGH); // Allume le rétroéclage de l'écran
    interval *= 2;   //Changement d'échelle du graphique                                       
    if (interval > 8) interval = 1;
      set_period = BASE_PERIOD * interval;   //Recalculer l'intervalle de mise à jour des données
    get_data ();      //Et mettre à jour les informations sur l'écran
  }
                            
  if (millis() - tmr1 >= BASE_PERIOD) {  //Collectons le tableau de données de base
    tmr1 = millis();   //Toutes les 11,25 minutes, nous prenons des mesures
    for (int i = 126; i >= 0; i--) {    //Pour écaler l'ensemble du tableau d'un pas
      base_array[i + 1] = base_array[i];
    }
value = round (analogRead(A0))*5/10.23;   //Résolution sur 5 volts / 1023 pas * tension lue * 100 = valeur
    base_array[0] = value;     //Écriture les dernières lectures dans un tableau
    update_all ();        //Et s'en souvenir dans l'EEPROM
  }

  if (millis() - tmr2 >= set_period) {  //Après une période de temps définie ( en millisecondes)
    tmr2 = millis();   //Actualisation des données à l'écran
    get_data ();
  }
}

void get_data () {    //Traçage des 16 colonnes du diagramme sur l'écran
  for (int i = 15; i >= 0; i--) {
    drawPlot(0, 3, 16, 4, MIN_VAL, MAX_VAL, (base_array[i * interval]));
  }
  delta = ((base_array[0]) - (base_array[15 * interval]));  //Calcul de l'écart (changement de tension) pour l'intervalle sélectionné
  screen_data (value, delta, (interval * 3));       //Et affichage d'informations du texte sur l'écran
}

void screen_data (int value, int delta, byte interval) {    //Affichage d'informations de texte sur l'écran (tout sauf les barres graphiques)
  lcd.setCursor(17, 0);
  lcd.print(value);
  lcd.setCursor(16, 2);
  if (delta == value) delta = 0;
  if (delta > 0) {
    lcd.print("+");
    digitalWrite(LedPin, HIGH); // Allume le rétroéclage de l'écran
  } else if (delta < 0) {
    lcd.print("-");
    digitalWrite(LedPin, HIGH); // Allume le rétroéclage de l'écran
  } else if (delta == 0) {
    lcd.print("    ");
    delay (5000);
    digitalWrite(LedPin, LOW); // Eteind le rétroéclage de l'écran
  }
  lcd.setCursor(17, 2);
  lcd.print(abs(delta));
  if (abs(delta) < 10) {
    lcd.setCursor(19, 2);
    lcd.print(" ");
  }
  lcd.setCursor(16, 1);
  lcd.print("Volt");
  lcd.setCursor(17, 3);
  lcd.print(interval);
  (interval < 10) ? lcd.print("h ") : lcd.print("h");
}

void initPlot() {
  // Symbols pour graphique : http://maxpromer.github.io/LCD-Character-Creator/
  byte row8[8] = {0b11111,  0b11111,  0b11111,  0b11111,  0b11111,  0b11111,  0b11111,  0b11111};
  byte row7[8] = {0b00000,  0b11111,  0b11111,  0b11111,  0b11111,  0b11111,  0b11111,  0b11111};
  byte row6[8] = {0b00000,  0b00000,  0b11111,  0b11111,  0b11111,  0b11111,  0b11111,  0b11111};
  byte row5[8] = {0b00000,  0b00000,  0b00000,  0b11111,  0b11111,  0b11111,  0b11111,  0b11111};
  byte row4[8] = {0b00000,  0b00000,  0b00000,  0b00000,  0b11111,  0b11111,  0b11111,  0b11111};
  byte row3[8] = {0b00000,  0b00000,  0b00000,  0b00000,  0b00000,  0b11111,  0b11111,  0b11111};
  byte row2[8] = {0b00000,  0b00000,  0b00000,  0b00000,  0b00000,  0b00000,  0b11111,  0b11111};
  byte row1[8] = {0b00000,  0b00000,  0b00000,  0b00000,  0b00000,  0b00000,  0b00000,  0b11111};
  lcd.createChar(0, row8);
  lcd.createChar(1, row1);
  lcd.createChar(2, row2);
  lcd.createChar(3, row3);
  lcd.createChar(4, row4);
  lcd.createChar(5, row5);
  lcd.createChar(6, row6);
  lcd.createChar(7, row7);
}
                                              
void drawPlot(byte pos, byte row, byte width, byte height, int min_val, int max_val, int fill_val) {
//Dessin de diagramme, code tiré de Guyver https://alexgyver.ru/lcd-plots-and-bars/
  for (byte i = 0; i < width; i++) {
    plot_array[i] = plot_array[i + 1];
  }
  fill_val = constrain(fill_val, min_val, max_val);
  plot_array[width - 1] = fill_val;
  for (byte i = 0; i < width; i++) { // Chaque colonne de paramètres
    int infill, fract;
  // Trouver le nombre de blocs entiers, en tenant compte du minimum et du maximum, à afficher sur le graphique                                            
    infill = floor((float)(plot_array[i] - min_val) / (max_val - min_val) * height * 10);
    fract = (infill % 10) * 8 / 10;  // Trouver le nombre de bandes restantes
    infill = infill / 12;
    for (byte n = 0; n < height; n++) {     // pour toutes les lignes du graphique
      if (n < infill && infill > 0) {  // tant que nous sommes en dessous du niveau
        lcd.setCursor(i, (row - n));  // remplir les cellules
        lcd.write((byte) 0);
      }
      if (n >= infill) {                          // Si le niveau est atteind
        lcd.setCursor(i, (row - n));
        if (fract > 0) lcd.write(fract);          // remplir des cellules fractionnaires
        else lcd.write(16);    // si les cellules fractionnaire == 0, remplir avec du vide
        for (byte k = n + 1; k < height; k++) {   // tout ce qui est au dessus est rempli de vide
          lcd.setCursor(i, (row - k));
          lcd.write(16);
        }
        break;
      }
    }
  }
}

void update_all () {   //Mettre à jour les données dans l'EEPROM
  eeprom_update_block((void*)&base_array, 0, sizeof(base_array));
}

void read_all () {   //Lecture des données de l'EEPROM 
  eeprom_read_block((void*)&base_array, 0, sizeof(base_array));
}